@@ -20,132 +20,31 @@ import poptorch
 from optimum.utils import logging
 from scipy.stats import truncnorm
 from transformers import (
+    BertForMaskedLM,
     BertForMultipleChoice,
     BertForPreTraining,
     BertForQuestionAnswering,
     BertForSequenceClassification,
     BertForTokenClassification,
 )
+from transformers.models.bert.modeling_bert import BertSelfAttention
 
-from ...modeling_utils import PipelineMixin, register
+from ...modeling_utils import (
+    OnehotGather,
+    PipelineMixin,
+    SerializedEmbedding,
+    SerializedLinear,
+    get_layer_ipu,
+    outline_attribute,
+    recomputation_checkpoint,
+    register,
+)
 from .bert_fused_attention import BertFusedSelfAttention
 
 
 logger = logging.get_logger(__name__)
 
 
-class OnehotGather(nn.Module):
-    """
-    Gathers selected indices from a tensor by transforming the list of indices
-    into a one-hot matrix and then multiplying the tensor by that matrix.
-    """
-
-    def __init__(self):
-        super().__init__()
-        self._is_half = False
-
-    def half(self):
-        super().half()
-        # Tracing is always executed in float as there are missing
-        # implementations of operations in half on the CPU.
-        # So we cannot query the inputs to know if we are running
-        # with a model that has had .half() called on it.
-        # To work around it nn.Module::half is overridden
-        self._is_half = True
-
-    def forward(self, sequence, positions):
-        """
-        Gather the vectors at the specific positions over a batch.
-        """
-        num_classes = int(sequence.shape[1])
-        one_hot_positions = F.one_hot(positions, num_classes)
-        if self._is_half:
-            one_hot_positions = one_hot_positions.half()
-        else:
-            one_hot_positions = one_hot_positions.float()
-        return torch.matmul(one_hot_positions.detach(), sequence)
-
-
-class SerializedLinear(nn.Linear):
-    """
-    Exactly equivalent to `nn.Linear` layer, but with the matrix multiplication replaced with
-    a serialized matrix multiplication: `poptorch.serializedMatMul`.
-    The matrix multiplication is split into separate smaller multiplications, calculated one after the other,
-    to reduce the memory requirements of the multiplication and its gradient calculation.
-
-    Args:
-        in_features: Size of each input sample
-        out_features: Size of each output sample
-        factor: Number of serialized multiplications. Must be a factor of
-            the dimension to serialize on.
-        bias: If set to False, the layer will not learn an additive bias.
-            Default: True
-        mode: Which dimension of the matmul to serialize on:
-            for matrix A (m by n) multiplied by matrix B (n by p).
-            * InputChannels: Split across the input channels (dimension m).
-            * ReducingDim: Split across the reducing dimension (n).
-            * OutputChannels: Split across the output channels (dimension p).
-            * Disabled: Same as an ordinary matrix multiplication.
-    """
-
-    def __init__(
-        self, in_features, out_features, factor, bias=False, mode=poptorch.MatMulSerializationMode.OutputChannels
-    ):
-        super().__init__(in_features, out_features, bias)
-        self.mode = mode
-        self.factor = factor
-
-    def forward(self, x):
-        output = poptorch.serializedMatMul(x, self.weight.t(), self.mode, self.factor)
-        if self.bias is not None:
-            output += self.bias
-        return output
-
-
-def _get_layer_ipu(layers_per_ipu):
-    # List of the IPU Id for each encoder layer
-    layer_ipu = []
-    for ipu, n_layers in enumerate(layers_per_ipu):
-        layer_ipu += [ipu] * n_layers
-    return layer_ipu
-
-
-def recomputation_checkpoint(module: nn.Module):
-    """Annotates the output of a module to be checkpointed instead of
-    recomputed"""
-
-    def recompute_outputs(module, inputs, outputs):
-        return tuple(poptorch.recomputationCheckpoint(y) for y in outputs)
-
-    return module.register_forward_hook(recompute_outputs)
-
-
-def outline_attribute(module: nn.Module, value: str):
-    """Adds an attribute to a module. This attribute will be used
-    when comparing operation equivalence in outlining. For example:
-
-    layer1 = nn.Linear(...)
-    layer2 = nn.Linear(...)
-    layer3 = nn.Linear(...)
-    layer4 = nn.Linear(...)
-    outline_attribute(layer1, "A")
-    outline_attribute(layer2, "A")
-    outline_attribute(layer3, "B")
-
-    The code for layer1 can be reused for layer2.
-    But it can't be used for layer3 or layer4.
-    """
-    context = poptorch.Attribute(__outline={"layer": value})
-
-    def enable(*args):
-        context.__enter__()
-
-    def disable(*args):
-        context.__exit__(None, None, None)
-
-    return module.register_forward_pre_hook(enable), module.register_forward_hook(disable)
-
-
 @register(BertForPreTraining)
 class PipelinedBertForPreTraining(BertForPreTraining, PipelineMixin):
     """
@@ -173,15 +72,13 @@ class PipelinedBertForPreTraining(BertForPreTraining, PipelineMixin):
 
         # Use faster fused-qkv self-attention
         for layer in self.bert.encoder.layer:
-            fused = BertFusedSelfAttention(self.config)
-            fused.load_state_dict(layer.attention.self.state_dict())
-            layer.attention.self = fused
+            layer.attention.self.__class__ = BertFusedSelfAttention
 
-        if self.config.embedding_serialization_factor > 1:
+        if self.ipu_config.embedding_serialization_factor > 1:
             serialized_decoder = SerializedLinear(
                 self.config.hidden_size,
                 self.config.vocab_size,
-                self.config.embedding_serialization_factor,
+                self.ipu_config.embedding_serialization_factor,
                 bias=True,
                 mode=poptorch.MatMulSerializationMode.OutputChannels,
             )
@@ -189,10 +86,10 @@ class PipelinedBertForPreTraining(BertForPreTraining, PipelineMixin):
             self.cls.predictions.decoder = serialized_decoder
             self.tie_weights()
 
-        layer_ipu = _get_layer_ipu(self.config.layers_per_ipu)
+        layer_ipu = get_layer_ipu(self.ipu_config.layers_per_ipu)
 
         logger.info("-------------------- Device Allocation --------------------")
-        logger.info("Embedding  --> IPU 0")
+        logger.info("Embedding --> IPU 0")
         self.bert.embeddings = poptorch.BeginBlock(self.bert.embeddings, "Embedding", ipu_id=0)
         # Preventing the embeddings.LayerNorm from being outlined with the encoder.layer.LayerNorm
         # improves the tile mapping of the pipeline stashes
@@ -201,13 +98,13 @@ class PipelinedBertForPreTraining(BertForPreTraining, PipelineMixin):
 
         for index, layer in enumerate(self.bert.encoder.layer):
             ipu = layer_ipu[index]
-            if self.config.recompute_checkpoint_every_layer:
+            if self.ipu_config.recompute_checkpoint_every_layer:
                 h = recomputation_checkpoint(layer)
                 self._hooks.append(h)
             self.bert.encoder.layer[index] = poptorch.BeginBlock(layer, f"Encoder{index}", ipu_id=ipu)
             logger.info(f"Encoder {index:<2} --> IPU {ipu}")
 
-        logger.info("Pooler     --> IPU 0")
+        logger.info("Pooler --> IPU 0")
         self.bert.pooler = poptorch.BeginBlock(self.bert.pooler, "Pooler", ipu_id=0)
 
         logger.info("Classifier --> IPU 0")
@@ -215,6 +112,28 @@ class PipelinedBertForPreTraining(BertForPreTraining, PipelineMixin):
         logger.info("-----------------------------------------------------------")
         return self
 
+    def deparallelize(self):
+        """
+        Undo the changes to the model done by `parallelize`.
+        You should call this before doing `save_pretrained` so that the `model.state_dict` is
+        compatible with the original model.
+        """
+        super().deparallelize()
+
+        for layer in self.bert.encoder.layer:
+            layer.attention.self.__class__ = BertSelfAttention
+
+        if self.ipu_config.embedding_serialization_factor > 1:
+            decoder = nn.Linear(
+                self.config.hidden_size,
+                self.config.vocab_size,
+                bias=True,
+            )
+            decoder.load_state_dict(self.cls.predictions.decoder.state_dict())
+            self.cls.predictions.decoder = decoder
+            self.tie_weights()
+        return self
+
     def _init_weights(self, module):
         """Initialize the weights"""
 
@@ -244,12 +163,6 @@ class PipelinedBertForPreTraining(BertForPreTraining, PipelineMixin):
         token_type_ids=None,
         labels=None,
         next_sentence_label=None,
-        output_attentions=None,
-        output_hidden_states=None,
-        return_dict=None,
-        position_ids=None,
-        head_mask=None,
-        inputs_embeds=None,
     ):
         output = self.bert(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
         sequence_output, pooled_output = output[:2]
@@ -284,68 +197,95 @@ class PipelinedBertForPreTraining(BertForPreTraining, PipelineMixin):
         return output
 
 
-class SerializedEmbedding(nn.Module):
+@register(BertForMaskedLM)
+class PipelinedBertForMaskedLM(BertForMaskedLM, PipelineMixin):
     """
-    Wrapper for `nn.Embedding` layer that performs the embedding look-up into
-    smaller serialized steps in order to reduce memory in the embedding gradient
-    calculation.
+    BertForMaskedLM transformed to run in an IPU pipeline.
 
-    Args:
-        embedding: A `nn.Embedding` to wrap
-        serialization_factor: The number of serialized embedding look-ups
+    Recommended usage:
+    ```
+    model = PipelinedBertForMaskedLM(config).parallelize().half().train()
+    ```
     """
 
-    def __init__(self, embedding: nn.Embedding, serialization_factor: int):
-        super().__init__()
-        self.serialization_factor = serialization_factor
-        self.num_embeddings = embedding.num_embeddings
+    def __init__(self, config):
+        super().__init__(config)
+        self.gather_indices = OnehotGather()
 
-        # Num embeddings should be divisible by the serialization factor
-        assert self.num_embeddings % self.serialization_factor == 0
-        self.split_size = self.num_embeddings // self.serialization_factor
-        self.split_embeddings = nn.ModuleList(
-            [
-                nn.Embedding.from_pretrained(
-                    embedding.weight[i * self.split_size : (i + 1) * self.split_size, :].detach(),
-                    freeze=False,
-                    padding_idx=embedding.padding_idx if i == 0 else None,
-                )
-                for i in range(self.serialization_factor)
-            ]
-        )
-
-    def deserialize(self):
+    def parallelize(self):
         """
-        Deserialize the internal wrapped embedding layer and return it as a
-        `nn.Embedding` object.
-
-        Returns:
-            `nn.Embedding` layer
+        Transform the model to run in an IPU pipeline.
+        - Adds pipeline stages to the model
+        - Replaces self-attention layers with fused-qkv self-attention layers
+        - (If enabled) Replaces the word embedding projection with a SerializedLinear layer
+        - Adds recomputation checkpoints
         """
-        return nn.Embedding.from_pretrained(torch.vstack([l.weight for l in self.split_embeddings]), padding_idx=0)
+        super().parallelize()
 
-    def forward(self, indices):
-        # iterate through the splits
-        x_sum = None
-        for i in range(self.serialization_factor):
-            # mask out the indices not in this split
-            split_indices = indices - i * self.split_size
-            mask = (split_indices >= 0) * (split_indices < self.split_size)
-            mask = mask.detach()
-            split_indices *= mask
+        # Use faster fused-qkv self-attention
+        for layer in self.bert.encoder.layer:
+            fused = BertFusedSelfAttention(self.config)
+            fused.load_state_dict(layer.attention.self.state_dict())
+            layer.attention.self = fused
 
-            # do the embedding lookup
-            x = self.split_embeddings[i](split_indices)
+        if self.ipu_config.embedding_serialization_factor > 1:
+            serialized_decoder = SerializedLinear(
+                self.config.hidden_size,
+                self.config.vocab_size,
+                self.ipu_config.embedding_serialization_factor,
+                bias=True,
+                mode=poptorch.MatMulSerializationMode.OutputChannels,
+            )
+            serialized_decoder.load_state_dict(self.cls.predictions.decoder.state_dict())
+            self.cls.predictions.decoder = serialized_decoder
+            self.tie_weights()
 
-            # multiply the output by mask
-            x *= mask.unsqueeze(-1)
+        layer_ipu = get_layer_ipu(self.ipu_config.layers_per_ipu)
 
-            # add to partial
-            if x_sum is not None:
-                x_sum += x
-            else:
-                x_sum = x
-        return x_sum
+        logger.info("-------------------- Device Allocation --------------------")
+        logger.info("Embedding  --> IPU 0")
+        self.bert.embeddings = poptorch.BeginBlock(self.bert.embeddings, "Embedding", ipu_id=0)
+        # Preventing the embeddings.LayerNorm from being outlined with the encoder.layer.LayerNorm
+        # improves the tile mapping of the pipeline stashes
+        hs = outline_attribute(self.bert.embeddings.LayerNorm, "embeddings")
+        self._hooks.extend(hs)
+
+        for index, layer in enumerate(self.bert.encoder.layer):
+            ipu = layer_ipu[index]
+            if self.ipu_config.recompute_checkpoint_every_layer:
+                h = recomputation_checkpoint(layer)
+                self._hooks.append(h)
+            self.bert.encoder.layer[index] = poptorch.BeginBlock(layer, f"Encoder{index}", ipu_id=ipu)
+            logger.info(f"Encoder {index:<2} --> IPU {ipu}")
+
+        logger.info("Classifier --> IPU 0")
+        self.cls = poptorch.BeginBlock(self.cls, "Classifier", ipu_id=0)
+        logger.info("-----------------------------------------------------------")
+        return self
+
+    def forward(self, input_ids, attention_mask, token_type_ids=None, labels=None):
+        output = self.bert(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
+        sequence_output = output[0]
+
+        if labels is not None:
+            # Select only the masked tokens for the classifier
+            max_number_of_masked_tokens = int(labels.size(1) * 0.25)
+            masked_lm_labels, masked_lm_positions = torch.topk(labels, k=max_number_of_masked_tokens, dim=1)
+            masked_output = self.gather_indices(sequence_output, masked_lm_positions)
+        else:
+            # This case should never happen during training
+            masked_output = sequence_output
+
+        prediction_scores = self.cls(masked_output)
+        output = (prediction_scores,) + output[2:]
+
+        if labels is not None:
+            masked_lm_loss = F.cross_entropy(
+                prediction_scores.view(-1, self.config.vocab_size), masked_lm_labels.view(-1)
+            ).float()
+            return masked_lm_loss
+
+        return output
 
 
 class BertPipelineMixin(PipelineMixin):
@@ -361,17 +301,15 @@ class BertPipelineMixin(PipelineMixin):
 
         # Use faster fused-qkv self-attention
         for layer in self.bert.encoder.layer:
-            fused = BertFusedSelfAttention(self.config)
-            fused.load_state_dict(layer.attention.self.state_dict())
-            layer.attention.self = fused
+            layer.attention.self.__class__ = BertFusedSelfAttention
 
-        layer_ipu = _get_layer_ipu(self.config.layers_per_ipu)
+        layer_ipu = get_layer_ipu(self.ipu_config.layers_per_ipu)
 
         logger.info("-------------------- Device Allocation --------------------")
-        logger.info("Embedding  --> IPU 0")
-        if self.config.embedding_serialization_factor > 1:
+        logger.info("Embedding --> IPU 0")
+        if self.ipu_config.embedding_serialization_factor > 1:
             self.bert.embeddings.word_embeddings = SerializedEmbedding(
-                self.bert.embeddings.word_embeddings, self.config.embedding_serialization_factor
+                self.bert.embeddings.word_embeddings, self.ipu_config.embedding_serialization_factor
             )
         self.bert.embeddings = poptorch.BeginBlock(self.bert.embeddings, "Embedding", ipu_id=0)
         hs = outline_attribute(self.bert.embeddings.LayerNorm, "embedding")
@@ -379,7 +317,7 @@ class BertPipelineMixin(PipelineMixin):
 
         for index, layer in enumerate(self.bert.encoder.layer):
             ipu = layer_ipu[index]
-            if self.config.recompute_checkpoint_every_layer and index != self.config.num_hidden_layers - 1:
+            if self.ipu_config.recompute_checkpoint_every_layer and index != self.config.num_hidden_layers - 1:
                 h = recomputation_checkpoint(layer)
                 self._hooks.append(h)
             self.bert.encoder.layer[index] = poptorch.BeginBlock(layer, f"Encoder{index}", ipu_id=ipu)
@@ -393,8 +331,12 @@ class BertPipelineMixin(PipelineMixin):
         compatible with the original model.
         """
         super().deparallelize()
+
+        for layer in self.bert.encoder.layer:
+            layer.attention.self.__class__ = BertSelfAttention
+
         # Deserialize the serialized word embedding
-        if self.config.embedding_serialization_factor > 1:
+        if self.ipu_config.embedding_serialization_factor > 1:
             self.bert.embeddings.word_embeddings = self.bert.embeddings.word_embeddings.deserialize()
         return self
 
@@ -412,7 +354,7 @@ class PipelinedBertForSequenceClassification(BertForSequenceClassification, Bert
 
     def parallelize(self):
         super().parallelize()
-        last_ipu = self.config.ipus_per_replica - 1
+        last_ipu = self.ipu_config.ipus_per_replica - 1
         logger.info(f"Classifier Output --> IPU {last_ipu}")
         self.classifier = poptorch.BeginBlock(self.classifier, "Classifier Output", ipu_id=last_ipu)
         logger.info("-----------------------------------------------------------")
@@ -441,7 +383,7 @@ class PipelinedBertForMultipleChoice(BertForMultipleChoice, BertPipelineMixin):
 
     def parallelize(self):
         super().parallelize()
-        last_ipu = self.config.ipus_per_replica - 1
+        last_ipu = self.ipu_config.ipus_per_replica - 1
         logger.info(f"Classifier Output --> IPU {last_ipu}")
         self.classifier = poptorch.BeginBlock(self.classifier, "Classifier Output", ipu_id=last_ipu)
         logger.info("-----------------------------------------------------------")
@@ -470,7 +412,7 @@ class PipelinedBertForTokenClassification(BertForTokenClassification, BertPipeli
 
     def parallelize(self):
         super().parallelize()
-        last_ipu = self.config.ipus_per_replica - 1
+        last_ipu = self.ipu_config.ipus_per_replica - 1
         logger.info(f"Classifier Output --> IPU {last_ipu}")
         self.classifier = poptorch.BeginBlock(self.classifier, "Classifier Output", ipu_id=last_ipu)
         logger.info("-----------------------------------------------------------")
@@ -499,7 +441,7 @@ class PipelinedBertForQuestionAnswering(BertForQuestionAnswering, BertPipelineMi
 
     def parallelize(self):
         super().parallelize()
-        last_ipu = self.config.ipus_per_replica - 1
+        last_ipu = self.ipu_config.ipus_per_replica - 1
         logger.info(f"QA Outputs --> IPU {last_ipu}")
         self.qa_outputs = poptorch.BeginBlock(self.qa_outputs, "QA Outputs", ipu_id=last_ipu)
         logger.info("-----------------------------------------------------------")
