@@ -13,131 +13,33 @@
 # limitations under the License.
 
 import torch
-import torch.nn as nn
+import torch.nn.functional as F
 
 import poptorch
 from optimum.utils import logging
 from transformers import (
+    RobertaForMaskedLM,
     RobertaForMultipleChoice,
     RobertaForQuestionAnswering,
     RobertaForSequenceClassification,
     RobertaForTokenClassification,
 )
 
-from ...modeling_utils import PipelineMixin, register
+from ...modeling_utils import (
+    OnehotGather,
+    PipelineMixin,
+    SerializedEmbedding,
+    SerializedLinear,
+    get_layer_ipu,
+    outline_attribute,
+    recomputation_checkpoint,
+    register,
+)
 
 
 logger = logging.get_logger(__name__)
 
 
-def _get_layer_ipu(layers_per_ipu):
-    # List of the IPU Id for each encoder layer
-    layer_ipu = []
-    for ipu, n_layers in enumerate(layers_per_ipu):
-        layer_ipu += [ipu] * n_layers
-    return layer_ipu
-
-
-def recomputation_checkpoint(module: nn.Module):
-    """Annotates the output of a module to be checkpointed instead of
-    recomputed"""
-
-    def recompute_outputs(module, inputs, outputs):
-        return tuple(poptorch.recomputationCheckpoint(y) for y in outputs)
-
-    return module.register_forward_hook(recompute_outputs)
-
-
-def outline_attribute(module: nn.Module, value: str):
-    """Adds an attribute to a module. This attribute will be used
-    when comparing operation equivalence in outlining. For example:
-
-    layer1 = nn.Linear(...)
-    layer2 = nn.Linear(...)
-    layer3 = nn.Linear(...)
-    layer4 = nn.Linear(...)
-    outline_attribute(layer1, "A")
-    outline_attribute(layer2, "A")
-    outline_attribute(layer3, "B")
-
-    The code for layer1 can be reused for layer2.
-    But it can't be used for layer3 or layer4.
-    """
-    context = poptorch.Attribute(__outline={"layer": value})
-
-    def enable(*args):
-        context.__enter__()
-
-    def disable(*args):
-        context.__exit__(None, None, None)
-
-    return module.register_forward_pre_hook(enable), module.register_forward_hook(disable)
-
-
-class SerializedEmbedding(nn.Module):
-    """
-    Wrapper for `nn.Embedding` layer that performs the embedding look-up into
-    smaller serialized steps in order to reduce memory in the embedding gradient
-    calculation.
-
-    Args:
-        embedding: A `nn.Embedding` to wrap
-        serialization_factor: The number of serialized embedding look-ups
-    """
-
-    def __init__(self, embedding: nn.Embedding, serialization_factor: int):
-        super().__init__()
-        self.serialization_factor = serialization_factor
-        self.num_embeddings = embedding.num_embeddings
-
-        # Num embeddings should be divisible by the serialization factor
-        assert self.num_embeddings % self.serialization_factor == 0
-        self.split_size = self.num_embeddings // self.serialization_factor
-        self.split_embeddings = nn.ModuleList(
-            [
-                nn.Embedding.from_pretrained(
-                    embedding.weight[i * self.split_size : (i + 1) * self.split_size, :].detach(),
-                    freeze=False,
-                    padding_idx=embedding.padding_idx if i == 0 else None,
-                )
-                for i in range(self.serialization_factor)
-            ]
-        )
-
-    def deserialize(self):
-        """
-        Deserialize the internal wrapped embedding layer and return it as a
-        `nn.Embedding` object.
-
-        Returns:
-            `nn.Embedding` layer
-        """
-        return nn.Embedding.from_pretrained(torch.vstack([l.weight for l in self.split_embeddings]), padding_idx=0)
-
-    def forward(self, indices):
-        # iterate through the splits
-        x_sum = None
-        for i in range(self.serialization_factor):
-            # mask out the indices not in this split
-            split_indices = indices - i * self.split_size
-            mask = (split_indices >= 0) * (split_indices < self.split_size)
-            mask = mask.detach()
-            split_indices *= mask
-
-            # do the embedding lookup
-            x = self.split_embeddings[i](split_indices)
-
-            # multiply the output by mask
-            x *= mask.unsqueeze(-1)
-
-            # add to partial
-            if x_sum is not None:
-                x_sum += x
-            else:
-                x_sum = x
-        return x_sum
-
-
 class RobertaPipelineMixin(PipelineMixin):
     def parallelize(self):
         """
@@ -147,13 +49,13 @@ class RobertaPipelineMixin(PipelineMixin):
         - Adds recomputation checkpoints
         """
         super().parallelize()
-        layer_ipu = _get_layer_ipu(self.config.layers_per_ipu)
+        layer_ipu = get_layer_ipu(self.ipu_config.layers_per_ipu)
 
         logger.info("-------------------- Device Allocation --------------------")
         logger.info("Embedding  --> IPU 0")
-        if self.config.embedding_serialization_factor > 1:
+        if self.ipu_config.embedding_serialization_factor > 1:
             self.roberta.embeddings.word_embeddings = SerializedEmbedding(
-                self.roberta.embeddings.word_embeddings, self.config.embedding_serialization_factor
+                self.roberta.embeddings.word_embeddings, self.ipu_config.embedding_serialization_factor
             )
         self.roberta.embeddings = poptorch.BeginBlock(self.roberta.embeddings, "Embedding", ipu_id=0)
         hs = outline_attribute(self.roberta.embeddings.LayerNorm, "embedding")
@@ -161,7 +63,7 @@ class RobertaPipelineMixin(PipelineMixin):
 
         for index, layer in enumerate(self.roberta.encoder.layer):
             ipu = layer_ipu[index]
-            if self.config.recompute_checkpoint_every_layer and index != self.config.num_hidden_layers - 1:
+            if self.ipu_config.recompute_checkpoint_every_layer and index != self.config.num_hidden_layers - 1:
                 h = recomputation_checkpoint(layer)
                 self._hooks.append(h)
             self.roberta.encoder.layer[index] = poptorch.BeginBlock(layer, f"Encoder{index}", ipu_id=ipu)
@@ -176,11 +78,93 @@ class RobertaPipelineMixin(PipelineMixin):
         """
         super().deparallelize()
         # Deserialize the serialized word embedding
-        if self.config.embedding_serialization_factor > 1:
+        if self.ipu_config.embedding_serialization_factor > 1:
             self.roberta.embeddings.word_embeddings = self.roberta.embeddings.word_embeddings.deserialize()
         return self
 
 
+@register(RobertaForMaskedLM)
+class PipelinedRobertaForMaskedLM(RobertaForMaskedLM, PipelineMixin):
+    """
+    RobertaForMaskedLM transformed to run in an IPU pipeline.
+
+    Recommended usage:
+    ```
+    model = PipelinedRobertaForMaskedLM(config).parallelize().half()
+    ```
+    """
+
+    def __init__(self, config):
+        super().__init__(config)
+        self.gather_indices = OnehotGather()
+
+    def parallelize(self):
+        """
+        Transform the model to run in an IPU pipeline.
+        - Adds pipeline stages to the model
+        - (If enabled) Replaces the word embedding projection with a SerializedLinear layer
+        - Adds recomputation checkpoints
+        """
+        super().parallelize()
+
+        if self.ipu_config.embedding_serialization_factor > 1:
+            serialized_decoder = SerializedLinear(
+                self.config.hidden_size,
+                self.config.vocab_size,
+                self.ipu_config.embedding_serialization_factor,
+                bias=True,
+                mode=poptorch.MatMulSerializationMode.OutputChannels,
+            )
+            serialized_decoder.load_state_dict(self.lm_head.decoder.state_dict())
+            self.lm_head.decoder = serialized_decoder
+            self.tie_weights()
+
+        layer_ipu = get_layer_ipu(self.ipu_config.layers_per_ipu)
+
+        logger.info("-------------------- Device Allocation --------------------")
+        logger.info("Embedding  --> IPU 0")
+        self.roberta.embeddings = poptorch.BeginBlock(self.roberta.embeddings, "Embedding", ipu_id=0)
+        hs = outline_attribute(self.roberta.embeddings.LayerNorm, "embedding")
+        self._hooks.extend(hs)
+
+        for index, layer in enumerate(self.roberta.encoder.layer):
+            ipu = layer_ipu[index]
+            if self.ipu_config.recompute_checkpoint_every_layer:
+                h = recomputation_checkpoint(layer)
+                self._hooks.append(h)
+            self.roberta.encoder.layer[index] = poptorch.BeginBlock(layer, f"Encoder{index}", ipu_id=ipu)
+            logger.info(f"Encoder {index:<2} --> IPU {ipu}")
+
+        logger.info("LM Head    --> IPU 0")
+        self.lm_head = poptorch.BeginBlock(self.lm_head, "LM Head", ipu_id=0)
+        logger.info("-----------------------------------------------------------")
+        return self
+
+    def forward(self, input_ids, attention_mask, labels=None):
+        outputs = self.roberta(input_ids, attention_mask=attention_mask)
+        sequence_output = outputs[0]
+
+        if labels is not None:
+            # Select only the masked tokens for the classifier
+            max_number_of_masked_tokens = int(labels.size(1) * 0.25)
+            masked_lm_labels, masked_lm_positions = torch.topk(labels, k=max_number_of_masked_tokens, dim=1)
+            masked_output = self.gather_indices(sequence_output, masked_lm_positions)
+        else:
+            # This case should never happen during training
+            masked_output = sequence_output
+
+        prediction_scores = self.lm_head(masked_output)
+        outputs = (prediction_scores,) + outputs[2:]
+
+        if labels is not None:
+            masked_lm_loss = F.cross_entropy(
+                prediction_scores.view(-1, self.config.vocab_size), masked_lm_labels.view(-1)
+            ).float()
+            return masked_lm_loss
+
+        return outputs
+
+
 @register(RobertaForSequenceClassification)
 class PipelinedRobertaForSequenceClassification(RobertaForSequenceClassification, RobertaPipelineMixin):
     """
@@ -194,7 +178,7 @@ class PipelinedRobertaForSequenceClassification(RobertaForSequenceClassification
 
     def parallelize(self):
         super().parallelize()
-        last_ipu = self.config.ipus_per_replica - 1
+        last_ipu = self.ipu_config.ipus_per_replica - 1
         logger.info(f"Classifier Output --> IPU {last_ipu}")
         self.classifier = poptorch.BeginBlock(self.classifier, "Classifier Output", ipu_id=last_ipu)
         logger.info("-----------------------------------------------------------")
@@ -222,7 +206,7 @@ class PipelinedRobertaForMultipleChoice(RobertaForMultipleChoice, RobertaPipelin
 
     def parallelize(self):
         super().parallelize()
-        last_ipu = self.config.ipus_per_replica - 1
+        last_ipu = self.ipu_config.ipus_per_replica - 1
         logger.info(f"Classifier Output --> IPU {last_ipu}")
         self.classifier = poptorch.BeginBlock(self.classifier, "Classifier Output", ipu_id=last_ipu)
         logger.info("-----------------------------------------------------------")
@@ -250,7 +234,7 @@ class PipelinedRobertaForTokenClassification(RobertaForTokenClassification, Robe
 
     def parallelize(self):
         super().parallelize()
-        last_ipu = self.config.ipus_per_replica - 1
+        last_ipu = self.ipu_config.ipus_per_replica - 1
         logger.info(f"Classifier Output --> IPU {last_ipu}")
         self.classifier = poptorch.BeginBlock(self.classifier, "Classifier Output", ipu_id=last_ipu)
         logger.info("-----------------------------------------------------------")
@@ -278,7 +262,7 @@ class PipelinedRobertaForQuestionAnswering(RobertaForQuestionAnswering, RobertaP
 
     def parallelize(self):
         super().parallelize()
-        last_ipu = self.config.ipus_per_replica - 1
+        last_ipu = self.ipu_config.ipus_per_replica - 1
         logger.info(f"QA Outputs --> IPU {last_ipu}")
         self.qa_outputs = poptorch.BeginBlock(self.qa_outputs, "QA Outputs", ipu_id=last_ipu)
         logger.info("-----------------------------------------------------------")
